/* -*-  Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1999  International Computer Science Institute
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by ACIRI, the AT&T 
 *      Center for Internet Research at ICSI (the International Computer
 *      Science Institute).
 * 4. Neither the name of ACIRI nor of ICSI may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <sys/types.h>
#include <math.h>
 
#include "tfrc_psnr.h"
#include "flags.h"

static class TFRCPSNRHeaderClass : public PacketHeaderClass {
public:
	TFRCPSNRHeaderClass() : PacketHeaderClass("PacketHeader/TFRCPSNR",
					      sizeof(hdr_tfrc)) {
		bind_offset(&hdr_tfrc::offset_);
	}
} class_tfrchdr_psnr;

static class TFRCPSNR_ACKHeaderClass : public PacketHeaderClass {
public:
	TFRCPSNR_ACKHeaderClass() : PacketHeaderClass("PacketHeader/TFRCPSNR_ACK",
						  sizeof(hdr_tfrc_ack)) {
		bind_offset(&hdr_tfrc_ack::offset_);
	}
} class_tfrc_psnr_ackhdr;

static class TfrcPSNRClass : public TclClass {
public:
  	TfrcPSNRClass() : TclClass("Agent/TFRCPSNR") {}
  	TclObject* create(int, const char*const*) {
    		return (new TfrcPSNRAgent());
  	}
} class_tfrc_psnr;


TfrcPSNRAgent::TfrcPSNRAgent() : Agent(PT_TFRC), send_timer_(this), 
	 NoFeedbacktimer_(this), rate_(0), oldrate_(0), maxrate_(0)
{
	bind("packetSize_", &size_);
	bind("rate_", &rate_);
	bind("df_", &df_);
	bind("tcp_tick_", &tcp_tick_);
	bind("ndatapack_", &ndatapack_);
	bind("ndatabytes_", &ndatabytes_);
	bind("true_loss_rate_", &true_loss_rate_);
	bind("srtt_init_", &srtt_init_);
	bind("rttvar_init_", &rttvar_init_);
	bind("rtxcur_init_", &rtxcur_init_);
	bind("rttvar_exp_", &rttvar_exp_);
	bind("T_SRTT_BITS", &T_SRTT_BITS);
	bind("T_RTTVAR_BITS", &T_RTTVAR_BITS);
	bind("InitRate_", &InitRate_);
	bind("overhead_", &overhead_);
	bind("ssmult_", &ssmult_);
	bind("bval_", &bval_);
	bind("ca_", &ca_);
	bind_bool("printStatus_", &printStatus_);
	bind_bool("conservative_", &conservative_);
	bind_bool("ecn_", &ecn_);
	bind("minrto_", &minrto_);
	bind("maxHeavyRounds_", &maxHeavyRounds_);
	bind("SndrType_", &SndrType_); 
	bind("scmult_", &scmult_);
	bind_bool("oldCode_", &oldCode_);
	bind("rate_init_", &rate_init_);
	bind("rate_init_option_", &rate_init_option_);
	bind_bool("slow_increase_", &slow_increase_); 
	bind_bool("voip_", &voip_);
	bind("voip_max_pkt_rate_", &voip_max_pkt_rate_);
	bind("fsize_", &fsize_);
	bind_bool("useHeaders_", &useHeaders_);
	bind_bool("idleFix_", &idleFix_);
	bind("headersize_", &headersize_);
	seqno_ = -1;
	maxseq_ = 0;
	datalimited_ = 0;
	lastlimited_ = 0.0;
	last_pkt_time_ = 0.0;
	bind("maxqueue_", &maxqueue_);
	maxqueue_ = MAXSEQ;
	ai_psnr_ = 0.2;
	md_psnr_ = 0.9;
	min_psnr_ = 30.0;
	max_psnr_ = 45.0;
	tfrc_psnr_debug_ = 0;
	vid_profile_ = 1;
	decreased_last_time = 0;
	decreased_count = 5;
	count = 0;
	previous_congestion_event = 0;
	loss_flag = 0;
}

/*
 * Must convert bytes into packets. 
 * If nbytes == -1, this corresponds to infinite send.  We approximate
 * infinite by a very large number (MAXSEQ).
 * For simplicity, when bytes are converted to packets, fractional packets 
 * are always rounded up.  
 */
void TfrcPSNRAgent::sendmsg(int nbytes, const char* /*flags*/)
{
	double now = Scheduler::instance().clock();
	if (tfrc_psnr_debug_ == 1) printf("At %f sendmsg\n", now);
	if (nbytes == -1 && maxseq_ < MAXSEQ)
		advanceby(MAXSEQ - maxseq_);
        else if (size_ > 0) {
		int npkts = int(nbytes/size_);
		npkts += (nbytes%size_ ? 1 : 0);
		// maxqueue was added by Tom Phelan, to control the
		//   transmit queue from the application.
		if ((maxseq_ - seqno_) < maxqueue_) {
			advanceby(npkts);
		}
	}
}


void TfrcPSNRAgent::advanceby(int delta)
{
	double now = Scheduler::instance().clock();
	if (tfrc_psnr_debug_ == 1) printf("At %f advanceby\n", now);
  	maxseq_ += delta;
	
	if (seqno_ == -1) {
		// if no packets hve been sent so far, call start. 
  		start(); 
	} else if (datalimited_ && maxseq_ > seqno_) {
		// We were data-limited - send a packet now!
		// The old code always waited for a timer to expire!!
		datalimited_ = 0;
		lastlimited_ = Scheduler::instance().clock();
		all_idle_ = 0;
		if (!oldCode_) {
			sendpkt();
		}
	}
} 

int TfrcPSNRAgent::command(int argc, const char*const* argv)
{
	if (argc==2) {
		// are we an infinite sender?
		if ( (strcmp(argv[1],"start")==0) && (SndrType_ == 0)) {
			start();
			return TCL_OK;
		}
		if (strcmp(argv[1],"stop")==0) {
			stop();
			return TCL_OK;
		}
	}
  	if ((argc == 3) && (SndrType_ == 1)) {
		// or do we need an FTP type app? 
    	if (strcmp(argv[1], "advance") == 0) {
            	int newseq = atoi(argv[2]);
		// THIS ISN"T USED.
		// newseq: new sequence
		// seqno_: next sequence to be sent
		// maxseq_: max seq_  produced by app so far.
            	if (newseq > maxseq_)
                 	advanceby(newseq - maxseq_);
            	return (TCL_OK);
    	}
    	if (strcmp(argv[1], "advanceby") == 0) {
      		advanceby(atoi(argv[2]));
      		return (TCL_OK);
    		}
	}
  	if (argc == 3) {
		if (strcmp(argv[1], "VideoProfile") == 0) {
			vid_profile_ = atoi(argv[2]);
			int row, column;

			switch (vid_profile_) {
			case 1: for (row=0;row<RDO_SIZE;row++)
						for(column=0;column<2;column++)
							rate_psnr[row][column] = rate_psnr_city[row][column];
			        printf("Video profile = %d\n", vid_profile_);
				break;
			case 2: for (row=0;row<RDO_SIZE;row++)
						for(column=0;column<2;column++)
								rate_psnr[row][column] = rate_psnr_crew[row][column];
			        printf("Video profile = %d\n", vid_profile_);
				break;
			case 3: for (row=0;row<RDO_SIZE;row++)
						for(column=0;column<2;column++)
							rate_psnr[row][column] = rate_psnr_harbour[row][column];
			        printf("Video profile = %d\n", vid_profile_);
				break;
			case 4: for (row=0;row<RDO_SIZE;row++)
						for(column=0;column<2;column++)
							rate_psnr[row][column] = rate_psnr_soccer[row][column];
			        printf("Video profile = %d\n", vid_profile_);
				break;
			default:
				break;
			}

			for (row=0;row<111;row++)
				for(column=0;column<2;column++)
					rate_psnr_virtual[row][column] = rate_psnr_mean[row][column];
			return (TCL_OK);
		}
		if (strcmp(argv[1], "AI_PSNR") == 0) {
			ai_psnr_ = atof(argv[2]);
			printf("AI = %f\n", ai_psnr_);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "MD_PSNR") == 0) {
			md_psnr_ = atof(argv[2]);
			printf("MD = %f\n", md_psnr_);
			return (TCL_OK);
		}
  	}
	return (Agent::command(argc, argv));
}

void TfrcPSNRAgent::start()
{
	seqno_=0;				
	rate_ = InitRate_;
	delta_ = 0;
	oldrate_ = rate_;  
	rate_change_ = SLOW_START;
	UrgentFlag = 1;
	rtt_=0;	 
	sqrtrtt_=1;
	rttcur_=1;
	tzero_ = 0;
	last_change_=0;
	maxrate_ = 0; 
	ss_maxrate_ = 0;
	ndatapack_=0;
	ndatabytes_ = 0;
	true_loss_rate_ = 0;
	active_ = 1; 
	round_id = 0;
	heavyrounds_ = 0;
	t_srtt_ = int(srtt_init_/tcp_tick_) << T_SRTT_BITS;
	t_rttvar_ = int(rttvar_init_/tcp_tick_) << T_RTTVAR_BITS;
	t_rtxcur_ = rtxcur_init_;
	rcvrate = 0 ;
	all_idle_ = 0;

	first_pkt_rcvd = 0 ;
	// send the first packet
	sendpkt();
	// ... at initial rate
	send_timer_.resched(size_/rate_);
	// ... and start timer so we can cut rate 
	// in half if we do not get feedback
	NoFeedbacktimer_.resched(2*size_/rate_); 
}

void TfrcPSNRAgent::stop()
{
	active_ = 0;
	if (idleFix_) 
 		datalimited_ = 1;
	send_timer_.force_cancel();
}

void TfrcPSNRAgent::nextpkt()
{
	if (tfrc_psnr_debug_ == 1) {
	        double now = Scheduler::instance().clock();
                printf("At %f nextpkt\n", now);
        }

	double next = -1;
	double xrate = -1; 

	if (SndrType_ == 0) {
		sendpkt();
	}
	else {
		if (maxseq_ > seqno_)
			sendpkt();
		else
			datalimited_ = 1;
			
                if (debug_) {
		 	double now = Scheduler::instance().clock();
			printf("Time: %5.2f Datalimited now.\n", now);
		}
	}
	
	// If slow_increase_ is set, then during slow start, we increase rate
	// slowly - by amount delta per packet 
	if (slow_increase_ && round_id > 2 && (rate_change_ == SLOW_START) 
		       && (oldrate_+SMALLFLOAT< rate_)) {
		oldrate_ = oldrate_ + delta_;
		xrate = oldrate_;
	} else {
		if (ca_) {
			if (debug_) printf("SQRT: now: %5.2f factor: %5.2f\n", Scheduler::instance().clock(), sqrtrtt_/sqrt(rttcur_));
			xrate = rate_ * sqrtrtt_/sqrt(rttcur_);
		} else
			xrate = rate_;
	}
	if (xrate > SMALLFLOAT) {
		next = size_/xrate;
		if (voip_) {
	  	    	double min_interval = 1.0/voip_max_pkt_rate_;
		    	if (next < min_interval)
				next = min_interval;
		}
		//
		// randomize between next*(1 +/- woverhead_) 
		//
		next = next*(2*overhead_*Random::uniform()-overhead_+1);
		if (next > SMALLFLOAT)
			send_timer_.resched(next);
                else 
			send_timer_.resched(SMALLFLOAT);
	}
}

void TfrcPSNRAgent::update_rtt (double tao, double now) 
{
	if (tfrc_psnr_debug_ == 1) printf("At %f update rtt\n", now);
	/* the TCP update */
	t_rtt_ = int((now-tao) /tcp_tick_ + 0.5);
	if (t_rtt_==0) t_rtt_=1;
	if (t_srtt_ != 0) {
		register short rtt_delta;
		rtt_delta = t_rtt_ - (t_srtt_ >> T_SRTT_BITS);    
		if ((t_srtt_ += rtt_delta) <= 0)    
			t_srtt_ = 1;
		if (rtt_delta < 0)
			rtt_delta = -rtt_delta;
	  	rtt_delta -= (t_rttvar_ >> T_RTTVAR_BITS);
	  	if ((t_rttvar_ += rtt_delta) <= 0)  
			t_rttvar_ = 1;
	} else {
		t_srtt_ = t_rtt_ << T_SRTT_BITS;		
		t_rttvar_ = t_rtt_ << (T_RTTVAR_BITS-1);	
	}
	t_rtxcur_ = (((t_rttvar_ << (rttvar_exp_ + (T_SRTT_BITS - T_RTTVAR_BITS))) + t_srtt_)  >> T_SRTT_BITS ) * tcp_tick_;
	tzero_=t_rtxcur_;
 	if (tzero_ < minrto_) 
  		tzero_ = minrto_;

	/* fine grained RTT estimate for use in the equation */
	if (rtt_ > 0) {
		rtt_ = df_*rtt_ + ((1-df_)*(now - tao));
		sqrtrtt_ = df_*sqrtrtt_ + ((1-df_)*sqrt(now - tao));
	} else {
		rtt_ = now - tao;
		sqrtrtt_ = sqrt(now - tao);
	}
	rttcur_ = now - tao;
}

/*
 * Receive a status report from the receiver.
 */
void TfrcPSNRAgent::recv(Packet *pkt, Handler *)
{
	double now = Scheduler::instance().clock();
	if (tfrc_psnr_debug_ == 1) printf("At %f recv\n", now);
	hdr_tfrc_ack *nck = hdr_tfrc_ack::access(pkt);
	double ts = nck->timestamp_echo + nck->timestamp_offset;
	double rate_since_last_report = nck->rate_since_last_report;
	double flost = nck->flost;
	int losses = nck->losses;
	true_loss_rate_ = nck->true_loss;
	if (tfrc_psnr_debug_ == 1) printf("FLOST = %f losses = %d true_loss_rate = %f RATE = %f in kbps STATE = %d\n", flost, losses, nck->true_loss, rate_*8.0/1000.0, rate_change_);

	round_id ++ ;
	UrgentFlag = 0;

	if (round_id > 1 && rate_since_last_report > 0) {
		/* compute the max rate for slow-start as two times rcv rate */ 
		ss_maxrate_ = 2*rate_since_last_report*size_;
		if (conservative_) { 
			if (losses >= 1) {
				/* there was a loss in the most recent RTT */
				if (debug_) printf("time: %5.2f losses: %d rate %5.2f\n", 
				  now, losses, rate_since_last_report);
				maxrate_ = rate_since_last_report*size_;
			} else { 
				/* there was no loss in the most recent RTT */
				maxrate_ = scmult_*rate_since_last_report*size_;
			}
			if (debug_) printf("time: %5.2f losses: %d rate %5.2f maxrate: %5.2f\n", now, losses, rate_since_last_report, maxrate_);
		} else 
			maxrate_ = 2*rate_since_last_report*size_;
	} else {
		ss_maxrate_ = 0;
		maxrate_ = 0; 
	}
		
	/* update the round trip time */
	update_rtt (ts, now);

	/* .. and estimate of fair rate */
	if (voip_ != 1) {
		// From RFC 3714:
		// The voip flow gets to send at the same rate as
		//  a TCP flow with 1460-byte packets.
		fsize_ = size_;
	}
	if (tfrc_psnr_debug_) printf("%f Profile = %d flost = %f rtt_ = %f tzero_ = %f fsize_ = %d bval_ = %d\n", now, vid_profile_, flost, rtt_, tzero_, fsize_, bval_);
	rcvrate = p_to_b(flost, rtt_, tzero_, fsize_, bval_);
	if (tfrc_psnr_debug_) printf("EQUATION Profile = %d rcvrate = %f rate_ = %f\n", vid_profile_, rcvrate*8.0/1000.0, rate_*8.0/1000.0);
	// rcvrate is in bytes per second, based on fairness with a    
	// TCP connection with the same packet size size_.	      
	if (voip_) {
		// Subtract the bandwidth used by headers.
		double temp = rcvrate*(size_/(1.0*headersize_+size_));
		rcvrate = temp;
	}

	/* if we get no more feedback for some time, cut rate in half */
	double t = 2*rtt_ ; 
	if (t < 2*size_/rate_) 
		t = 2*size_/rate_ ; 
	NoFeedbacktimer_.resched(t);
	
	/* if we are in slow start and we just saw a loss */
	/* then come out of slow start */
	if (first_pkt_rcvd == 0) {
		first_pkt_rcvd = 1 ; 
		slowstart();
		nextpkt();
	}
	else {
		if (rate_change_ == SLOW_START) {
			if (flost > 0) {
				rate_change_ = OUT_OF_SLOW_START;
				oldrate_ = rate_ = rcvrate;
				if (tfrc_psnr_debug_ > 1) printf("At %f Profile = %d rate = %f = %f in kbps OUT OF SLOW START\n", now, vid_profile_, rate_, rate_*8.0/1000.0);
			}
			else {
				slowstart();
				nextpkt();
			}
		}
		else {
			/* Original TFRC code
			if (rcvrate>rate_) {
				if (tfrc_psnr_debug_) printf("Increase Rate CONGESTION AVOIDANCE\n");
				increase_rate(flost);
			}
			else
				decrease_rate ();
			*/
				double psnr = rate_to_psnr(rate_psnr, RDO_SIZE, rate_);
				virtual_rate_ = psnr_to_rate(rate_psnr_virtual, 111, psnr)*1000.0/8.0;
				if (rcvrate > virtual_rate_) {
					increase_rate_virtual_new(flost);
				} else {
					decrease_rate_virtual_new();
				}
		}
	}
	if (printStatus_) {
		printf("time: %5.2f rate: %5.2f\n", now, rate_);
		double packetrate = rate_ * rtt_ / size_;
		printf("time: %5.2f packetrate: %5.2f\n", now, packetrate);
		double maxrate = maxrate_ * rtt_ / size_;
		printf("time: %5.2f maxrate: %5.2f\n", now, maxrate);
	}
	Packet::free(pkt);
}

/*
 * Calculate initial sending rate from RFC 3390.
 */
double TfrcPSNRAgent::rfc3390(int size)
{
	if (tfrc_psnr_debug_ == 1) {
	        double now = Scheduler::instance().clock();
                printf("At %f rfc3390 size_ = %d\n", now, size_);
        }
	
        if (size_ <= 1095) {
                return (4.0);
        } else if (size_ < 2190) {
                return (3.0);
        } else {
                return (2.0);
        }
}

/*
 * Used in setting the initial rate.
 * This is from TcpAgent::initial_wnd().
 */
double TfrcPSNRAgent::initial_rate()
{
	if (tfrc_psnr_debug_ == 1) {
		double now = Scheduler::instance().clock();
                printf("At %f initial rate\n", now);
        }

	if (rate_init_option_ == 1) {
	// init_option = 1: static initial rate of rate_init_
               return (rate_init_);
        }
        else if (rate_init_option_ == 2) {
                // do initial rate according to RFC 3390.
		return (rfc3390(size_));
        }
        // XXX what should we return here???
        fprintf(stderr, "Wrong number of rate_init_option_ %d\n",
                rate_init_option_);
        abort();
        return (2.0); // XXX make msvc happy.
}


// ss_maxrate_ = 2*rate_since_last_report*size_;
// rate_: the rate set from the last pass through slowstart()
void TfrcPSNRAgent::slowstart () 
{
	double now = Scheduler::instance().clock();
	double initrate = initial_rate()*size_/rtt_;
	// If slow_increase_ is set to true, delta is used so that 
	//  the rate increases slowly to new value over an RTT. 
	if (debug_) printf("SlowStart: round_id: %d rate: %5.2f ss_maxrate_: %5.2f\n", round_id, rate_, ss_maxrate_);
	if (round_id <=1 || (round_id == 2 && initial_rate() > 1)) {
		// We don't have a good rate report yet, so keep to  
		//   the initial rate.				     
		oldrate_ = rate_;
		if (rate_ < initrate) rate_ = initrate;
		if (tfrc_psnr_debug_ > 1) printf("Profile = %d In slowstart 1111 rate_ = %f = %f in kbps\n", vid_profile_, rate_, rate_*8.0/1000.0);
		delta_ = (rate_ - oldrate_)/(rate_*rtt_/size_);
		last_change_=now;
	} else if (ss_maxrate_ > 0) {
		if (idleFix_ && (datalimited_ || lastlimited_ > now - 1.5*rtt_)
			     && ss_maxrate_ < initrate) {
			// Datalimited recently, and maxrate is small.
			// Don't be limited by maxrate to less that initrate.
			oldrate_ = rate_;
			if (rate_ < initrate) rate_ = initrate;
			if (tfrc_psnr_debug_ > 1) printf("Profile = %d In slowstart 2222 rate_ = %f = %f\n", vid_profile_, rate_, rate_*8.0/1000.0);
			delta_ = (rate_ - oldrate_)/(rate_*rtt_/size_);
			last_change_=now;
		} else if (rate_ < ss_maxrate_ && 
		                    now - last_change_ > rtt_) {
			// Not limited by maxrate, and time to increase.
			// Multiply the rate by ssmult_, if maxrate allows.
			oldrate_ = rate_;
			if (ssmult_*rate_ > ss_maxrate_) {
				rate_ = ss_maxrate_;
				if (tfrc_psnr_debug_ > 1) printf("Profile = %d In slowstart 3333 rate_ = %f = %f in kbps\n", vid_profile_, rate_, rate_*8.0/1000.0);
			}
			else rate_ = ssmult_*rate_;
			if (rate_ < size_/rtt_) {
				rate_ = size_/rtt_;
				if (tfrc_psnr_debug_ > 1) printf("Profile = %d In slowstart 4444 rate_ = %f = %f in kbps\n", vid_profile_, rate_, rate_*8.0/1000.0);
			}
			delta_ = (rate_ - oldrate_)/(rate_*rtt_/size_);
			last_change_=now;
		} else if (rate_ > ss_maxrate_) {
			// Limited by maxrate.  
			rate_ = oldrate_ = ss_maxrate_/2.0;
			if (tfrc_psnr_debug_ > 1) printf("Profile = %d In slowstart 5555 rate_ = %f = %f in kbps\n", vid_profile_, rate_, rate_*8.0/1000.0);
			delta_ = 0;
			last_change_=now;
		} 
	} else {
		// If we get here, ss_maxrate <= 0, so the receive rate is 0.
		// We should go back to a very small sending rate!!!
		oldrate_ = rate_;
		rate_ = size_/rtt_;
		if (tfrc_psnr_debug_ > 1) printf("Profile = %d In slowstart 6666 rate_ = %f = %f\n", vid_profile_, rate_, rate_*8.0/1000.0);
		delta_ = 0;
        	last_change_=now;
	}
	if (debug_) printf("SlowStart: now: %5.2f rate: %5.2f delta: %5.2f\n", now, rate_, delta_);
	if (printStatus_) {
		double rate = rate_ * rtt_ / size_;
	  	printf("SlowStart: now: %5.2f rate: %5.2f ss_maxrate: %5.2f delta: %5.2f\n", now, rate, ss_maxrate_, delta_);
	}
}

void TfrcPSNRAgent::increase_rate (double p)
{               
        double now = Scheduler::instance().clock();
		if (tfrc_psnr_debug_ == 1) printf("At %f INCREASE RATE\n", now);
        double maximumrate;

	double mult = (now-last_change_)/rtt_ ;
	if (mult > 2) mult = 2 ;
	if (tfrc_psnr_debug_ == 1) printf("At %f rate before = %f\n", now, rate_);
	rate_ = rate_ + (size_/rtt_)*mult ;
	if (tfrc_psnr_debug_ == 1) printf("At %f rate after = %f\n", now, rate_);
	if (datalimited_ || lastlimited_ > now - 1.5*rtt_) {
		// Modified by Sally on 3/10/2006
		// If the sender has been datalimited, rate should be
		//   at least the initial rate, when increasing rate.
		double init_rate = initial_rate()*size_/rtt_;
	   	maximumrate = (maxrate_>init_rate)?maxrate_:init_rate ;
        } else {
	   	maximumrate = (maxrate_>size_/rtt_)?maxrate_:size_/rtt_ ;
	}
	maximumrate = (maximumrate>rcvrate)?rcvrate:maximumrate;

	rate_ = (rate_ > maximumrate)?maximumrate:rate_ ;
	if (tfrc_psnr_debug_) printf("1111 At %f rate = %f\n", now, rate_);
	
        rate_change_ = CONG_AVOID;  
        last_change_ = now;
	heavyrounds_ = 0;
	if (debug_) printf("Increase: now: %5.2f rate: %5.2f lastlimited: %5.2f rtt: %5.2f\n", now, rate_, lastlimited_, rtt_);
	if (printStatus_) {
		double rate = rate_ * rtt_ / size_;
		printf("Increase: now: %5.2f rate: %5.2f lastlimited: %5.2f rtt: %5.2f\n", now, rate, lastlimited_, rtt_);
	}
	if (tfrc_psnr_debug_) printf("Increase: now: %5.2f rate: %5.2f lastlimited: %5.2f rtt: %5.2f\n", now, rate_, lastlimited_, rtt_);
}       

void TfrcPSNRAgent::increase_rate_virtual_new (double p)
{
	double now = Scheduler::instance().clock();
	double maximumrate;

	double mult = (now-last_change_)/rtt_ ;
	if (mult > 2) mult = 2 ;
	virtual_rate_ = virtual_rate_ + (size_/rtt_)*mult ;

	if (datalimited_ || lastlimited_ > now - 1.5*rtt_) {
		// Modified by Sally on 3/10/2006
		// If the sender has been datalimited, rate should be
		//   at least the initial rate, when increasing rate.
		double init_rate = initial_rate()*size_/rtt_;
	   	maximumrate = (maxrate_>init_rate)?maxrate_:init_rate ;
        } else {
	   	maximumrate = (maxrate_>size_/rtt_)?maxrate_:size_/rtt_ ;
	}
	maximumrate = (maximumrate>rcvrate)?rcvrate:maximumrate;

	rate_change_ = CONG_AVOID;
	last_change_ = now;
	heavyrounds_ = 0;


	double psnr;
	psnr = rate_to_psnr(rate_psnr_virtual, 111, virtual_rate_);
	rate_ = psnr_to_rate(rate_psnr, RDO_SIZE, psnr)*1000.0/8.0;
}

void TfrcPSNRAgent::increase_rate_virtual (double p)
{
	double now = Scheduler::instance().clock();
	double delta, rate_before, psnr, new_psnr, virtual_rate;
	double maximumrate;
	double mult = (now-last_change_)/rtt_ ;
	rate_before = rate_;
	printf("VIRTUAL rate_ = %f at begining of method rcvrate = %f\n", rate_*8.0/1000.0, rcvrate*8.00/1000.0);
	if (mult > 2) mult = 2 ;
	rate_ = rate_ + (size_/rtt_)*mult ;

	if (datalimited_ || lastlimited_ > now - 1.5*rtt_) {
		// Modified by Sally on 3/10/2006
		// If the sender has been datalimited, rate should be
		//   at least the initial rate, when increasing rate.
		double init_rate = initial_rate()*size_/rtt_;
	   	maximumrate = (maxrate_>init_rate)?maxrate_:init_rate ;
        } else {
	   	maximumrate = (maxrate_>size_/rtt_)?maxrate_:size_/rtt_ ;
	}
	maximumrate = (maximumrate>rcvrate)?rcvrate:maximumrate;
	printf("maximumrate = %f\n", maximumrate*8.0/1000.0);
	rate_ = (rate_ > maximumrate)?maximumrate:rate_ ;
	rate_change_ = CONG_AVOID;
	last_change_ = now;
	heavyrounds_ = 0;
	if (debug_) printf("Increase: now: %5.2f rate: %5.2f lastlimited: %5.2f rtt: %5.2f\n", now, rate_, lastlimited_, rtt_);
	if (printStatus_) {
		double rate = rate_ * rtt_ / size_;
		printf("Increase: now: %5.2f rate: %5.2f lastlimited: %5.2f rtt: %5.2f\n", now, rate, lastlimited_, rtt_);
	}

	delta = rate_ - rate_before; //in bytes
	printf("INCREASE delta = %f\n", delta*8.0/1000.0);
	psnr = rate_to_psnr(rate_psnr, RDO_SIZE, rate_before);
	virtual_rate = psnr_to_rate(rate_psnr_virtual, 111, psnr)*1000.0/8.0; //in bytes
	new_psnr = rate_to_psnr(rate_psnr_virtual, 111, virtual_rate+delta);
	rate_ = psnr_to_rate(rate_psnr, RDO_SIZE, new_psnr)*1000.0/8.0; //in bytes
}

void TfrcPSNRAgent::increase_rate_psnr (double p)
{
        double now = Scheduler::instance().clock();
		if (tfrc_psnr_debug_) printf("At %f INCREASE RATE PSNR\n", now);
        double maximumrate;

	double mult = (now-last_change_)/rtt_ ;
	if (mult > 2) mult = 2 ;
	if (tfrc_psnr_debug_ > 1) printf("At %f Profile = %d rate before MULT= %f = %f in kbps\n", now, vid_profile_, rate_, rate_*8.0/1000.0);
	if (tfrc_psnr_debug_ > 1) printf("At %f Profile = %d rate after MULT = %f = %f in kbps\n", now, vid_profile_, rate_, rate_*8.0/1000.0);
	if (datalimited_ || lastlimited_ > now - 1.5*rtt_) {
		// Modified by Sally on 3/10/2006
		// If the sender has been datalimited, rate should be
		//   at least the initial rate, when increasing rate.
		double init_rate = initial_rate()*size_/rtt_;
	   	maximumrate = (maxrate_>init_rate)?maxrate_:init_rate ;
        } else {
	   	maximumrate = (maxrate_>size_/rtt_)?maxrate_:size_/rtt_ ;
	}
	maximumrate = (maximumrate>rcvrate)?rcvrate:maximumrate;

	double psnr, rate_kbps;
	psnr = rate_to_psnr(rate_psnr, RDO_SIZE, rate_);

	//Increase rate only if psnr is smaller than MAX PSNR
	if (psnr < max_psnr_) {
		rate_kbps = psnr_to_rate(rate_psnr, RDO_SIZE, psnr+ai_psnr_);
		rate_ = (rate_kbps*1000.0)/8;
	}
	if (tfrc_psnr_debug_) printf("Profile = %d new RATE in kbps = %f in bytes = %f\n", vid_profile_, rate_kbps, rate_);

        rate_change_ = CONG_AVOID;
        last_change_ = now;
	heavyrounds_ = 0;
	if (debug_) printf("Increase: now: %5.2f rate: %5.2f lastlimited: %5.2f rtt: %5.2f\n", now, rate_, lastlimited_, rtt_);
	if (printStatus_) {
		double rate = rate_ * rtt_ / size_;
		printf("Increase: now: %5.2f rate: %5.2f lastlimited: %5.2f rtt: %5.2f\n", now, rate, lastlimited_, rtt_);
	}
}

double TfrcPSNRAgent::rate_to_psnr(double array[][2], int size, double r)
{
	int i;
	double psnr, slope;
	double rate_kbps;
	rate_kbps = r*8/1000.0;
	if (tfrc_psnr_debug_) printf("Profile = %d rate_kbps in RATE_TO_PSNR = %f\n", vid_profile_, rate_kbps);
	for(i=0;i<size;i++) {
		if (array[i][0] > rate_kbps) {
			if (i==0)
				return array[i][1]; //Return the smallest PSNR from the list
			slope = (array[i][1]-array[i-1][1])/ (array[i][0]-array[i-1][0]);
			//Calculate based on the closer reference point
			if ( rate_kbps < (array[i][0]+array[i-1][0])/2 )
				psnr = slope*(rate_kbps - array[i-1][0]) + array[i-1][1];
			else
				psnr = array[i][1] - slope*(array[i][0] - rate_kbps);
			break;
		}
	}
	if (tfrc_psnr_debug_) printf("Profile = %d Return PSNR = %f in RATE_TO_PSNR\n", vid_profile_, psnr);
	return psnr;
}

double TfrcPSNRAgent::psnr_to_rate(double array[][2], int size, double psnr)
{
	int i;
	double slope;
	double rate_kbps;
	if (tfrc_psnr_debug_) printf("Profile = %d psnr in PSNR_TO_RATE = %f\n", vid_profile_, psnr);
	for(i=0;i<size;i++) {
		if (array[i][1] > psnr) {
			if (i==0) {
				if (tfrc_psnr_debug_) printf("RETURN rate_kps = %f in bytes = %f when i =0\n", array[0][0], array[0][0]*1000.0/8);
				return array[0][0]*1000.0/8; //Return the smallest rate from the list
			}
			slope = (array[i][1]-array[i-1][1])/ (array[i][0]-array[i-1][0]);
			//Calculate based on the closer reference point
			if ( psnr < (array[i][1]+array[i-1][1])/2 ) {
				rate_kbps = (psnr - array[i-1][1])/slope + array[i-1][0];
			}
			else {
				rate_kbps = array[i][0] - (array[i][1] - psnr)/slope;
			}
			break;
		}
	}
	if (tfrc_psnr_debug_) printf("Profile = %d RETURN rate_kps in PSNR_TO_RATE = %f in kbps = %f in bytes\n", vid_profile_, rate_kbps, rate_kbps*1000.0/8);
	return rate_kbps; //in kbps
}

void TfrcPSNRAgent::decrease_rate () 
{
	double now = Scheduler::instance().clock(); 
	if (tfrc_psnr_debug_) printf("At %f decrease rate, rate = %f  = %f in kbps\n", now, rate_, rate_*8.0/1000.0);
	double psnr, rate_tmp;
	psnr = rate_to_psnr(rate_psnr, RDO_SIZE, rate_);
	if (tfrc_psnr_debug_) printf("PSNR = %f\n", psnr);
	rate_tmp = psnr_to_rate(rate_psnr, RDO_SIZE, psnr*md_psnr_);
	if (tfrc_psnr_debug_) printf("Reduced Rate PSNR to = %f in bytes, in kbps = %f\n", rate_tmp, rate_tmp*8/1000.0);
	if (tfrc_psnr_debug_) printf("1111 rate = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
	double maximumrate = (maxrate_>size_/rtt_)?maxrate_:size_/rtt_ ;

	// Allow sending rate to be greater than maximumrate
	//   (which is by default twice the receiving rate)
	//   for at most maxHeavyRounds_ rounds.
	if (rate_ > maximumrate)
		heavyrounds_++;
	else
		heavyrounds_ = 0;
	if (heavyrounds_ > maxHeavyRounds_) {
		rate_ = (rate_ > maximumrate)?maximumrate:rate_ ;
		if (tfrc_psnr_debug_) printf("2222 rate = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
	}

	rate_change_ = RATE_DECREASE;
	last_change_ = now;
	if (debug_) printf("Decrease: now: %5.2f rate: %5.2f rtt: %5.2f\n", now, rate_, rtt_);
	if (printStatus_) {
		double rate = rate_ * rtt_ / size_;
		printf("Decrease: now: %5.2f rate: %5.2f rtt: %5.2f\n", now, rate, rtt_);
	}
	if (tfrc_psnr_debug_) printf("Decrease: now: %5.2f rate: %5.2f rtt: %5.2f\n", now, rate_, rtt_);
}

void TfrcPSNRAgent::decrease_rate_virtual_new ()
{
	double now = Scheduler::instance().clock();
	virtual_rate_ = rcvrate;

	rate_change_ = RATE_DECREASE;
	last_change_ = now;

	double psnr;
	psnr = rate_to_psnr(rate_psnr_virtual, 111, virtual_rate_);
	rate_ = psnr_to_rate(rate_psnr, RDO_SIZE, psnr)*1000.0/8.0;
}

void TfrcPSNRAgent::decrease_rate_virtual ()
{
	double now = Scheduler::instance().clock();
	double delta, rate_before, psnr, new_psnr, virtual_rate;
	rate_before = rate_;
	rate_ = rcvrate;
	double maximumrate = (maxrate_>size_/rtt_)?maxrate_:size_/rtt_ ;

	// Allow sending rate to be greater than maximumrate
	//   (which is by default twice the receiving rate)
	//   for at most maxHeavyRounds_ rounds.
	if (rate_ > maximumrate)
		heavyrounds_++;
	else
		heavyrounds_ = 0;
	if (heavyrounds_ > maxHeavyRounds_) {
		rate_ = (rate_ > maximumrate)?maximumrate:rate_ ;
	}

	rate_change_ = RATE_DECREASE;
	last_change_ = now;
	if (debug_) printf("Decrease: now: %5.2f rate: %5.2f rtt: %5.2f\n", now, rate_, rtt_);
	if (printStatus_) {
		double rate = rate_ * rtt_ / size_;
		printf("Decrease: now: %5.2f rate: %5.2f rtt: %5.2f\n", now, rate, rtt_);
	}

	delta = rate_before - rate_; //in bytes
	printf("DECREASE delta = %f in kbps\n", delta*8.0/1000.0);
	psnr = rate_to_psnr(rate_psnr, RDO_SIZE, rate_before);
	virtual_rate = psnr_to_rate(rate_psnr_virtual, 111, psnr);
	printf("psnr = %f VIRTUAL RATE PURE %f\n", psnr, virtual_rate);
	virtual_rate = psnr_to_rate(rate_psnr_virtual, 111, psnr)*1000.0/8.0; //in bytes
	printf("DECREASE virtual_rate = %f = %f in kbps\n", virtual_rate, virtual_rate*8.0/1000.0);
	new_psnr = rate_to_psnr(rate_psnr_virtual, 111, virtual_rate-delta);
	printf("DECREASE new_psnr = %f\n", new_psnr);
	rate_ = psnr_to_rate(rate_psnr, RDO_SIZE, new_psnr)*1000.0/8.0; //in bytes
	printf("%f Profile = %d DECREASE Rate BEFORE = %f psnr = %f Rate AFTER = %f new_psnr = %f\n", now, vid_profile_, rate_before*8.0/1000.0, psnr, rate_*8.0/1000.0, new_psnr);
}

void TfrcPSNRAgent::decrease_rate_psnr ()
{
	double now = Scheduler::instance().clock();
	if (tfrc_psnr_debug_) printf("At %f decrease rate PSNR, rate = %f = %f in kbps\n", now, rate_, rate_*8.0/1000.0);
	double psnr, rate_kbps, delta_psnr, reduction;
	psnr = rate_to_psnr(rate_psnr, RDO_SIZE, rate_);
	//Decrease the rate only if current PSNR is greater to MIN PSNR

	if (psnr > min_psnr_) {
		delta_psnr = psnr - min_psnr_;
		reduction = delta_psnr * md_psnr_;
		rate_kbps = psnr_to_rate(rate_psnr, RDO_SIZE, min_psnr_ + reduction);
		rate_ = rate_kbps*1000.0/8.0;
	}

	double maximumrate = (maxrate_>size_/rtt_)?maxrate_:size_/rtt_ ;

	// Allow sending rate to be greater than maximumrate
	//   (which is by default twice the receiving rate)
	//   for at most maxHeavyRounds_ rounds.
	if (rate_ > maximumrate)
		heavyrounds_++;
	else
		heavyrounds_ = 0;
	if (heavyrounds_ > maxHeavyRounds_) {
		rate_ = (rate_ > maximumrate)?maximumrate:rate_ ;
		if (tfrc_psnr_debug_) printf("2222 rate = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
	}

	rate_change_ = RATE_DECREASE;
	last_change_ = now;
	if (debug_) printf("Decrease: now: %5.2f rate: %5.2f rtt: %5.2f\n", now, rate_, rtt_);
	if (printStatus_) {
		double rate = rate_ * rtt_ / size_;
		printf("Decrease: now: %5.2f rate: %5.2f rtt: %5.2f\n", now, rate, rtt_);
	}
}

void TfrcPSNRAgent::sendpkt()
{
	if (active_) {
		double now = Scheduler::instance().clock();
		if (tfrc_psnr_debug_) printf("At %f sendpkt\n", now);
		Packet* p = allocpkt();
		hdr_tfrc *tfrch = hdr_tfrc::access(p);
		hdr_flags* hf = hdr_flags::access(p);
		if (ecn_) {
			hf->ect() = 1;  // ECN-capable transport
		}
		tfrch->seqno=seqno_++;
		tfrch->timestamp=Scheduler::instance().clock();
		tfrch->rtt=rtt_;
		tfrch->tzero=tzero_;
		tfrch->rate=rate_;
		tfrch->psize=size_;
		tfrch->fsize=fsize_;
		tfrch->UrgentFlag=UrgentFlag;
		tfrch->round_id=round_id;
		ndatapack_++;
		ndatabytes_ += size_;
		if (useHeaders_ == true) {
			hdr_cmn::access(p)->size() += headersize_;
		}
		last_pkt_time_ = now;
	        send(p, 0);
	}
}


/*
 * RFC 3448:
 * "If the sender has been idle since this nofeedback timer was set and
 * X_recv is less than four packets per round-trip time, then X_recv
 * should not be halved in response to the timer expiration.  This
 * ensures that the allowed sending rate is never reduced to less than
 * two packets per round-trip time as a result of an idle period."
 */
 
void TfrcPSNRAgent::reduce_rate_on_no_feedback()
{
	double now = Scheduler::instance().clock();
	if (tfrc_psnr_debug_) printf("At %f reduce rate on no feedback\n", now);
	printf("At %f reduce rate on no feedback\n", now);
	if (rate_change_ != SLOW_START)
		// "if" statement added by Sally on 9/25/06,
		// so that an idle report doesn't kick us out of
		// slow-start!
		rate_change_ = RATE_DECREASE; 
	if (oldCode_ || (!all_idle_ && !datalimited_)) {
		// if we are not datalimited
		rate_*=0.5;
		if (tfrc_psnr_debug_) printf("1111 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
	} else if ((datalimited_ || all_idle_) && rate_init_option_ == 1) { 
		// all_idle_: the sender has been datalimited since the 
		//    timer was set
		//  Don't reduce rate below rate_init_ * size_/rtt_.
                if (rate_ > 2.0 * rate_init_ * size_/rtt_ ) {
                        rate_*=0.5;
                        if (tfrc_psnr_debug_) printf("2222 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
                } 
	} else if ((datalimited_ || all_idle_) && rate_init_option_ == 2) {
                // Don't reduce rate below the RFC3390 rate.
                if (rate_ > 2.0 * rfc3390(size_) * size_/rtt_ ) {
                        rate_*=0.5;
                        if (tfrc_psnr_debug_) printf("3333 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
                } else if ( rate_ > rfc3390(size_) * size_/rtt_ ) {
                        rate_ = rfc3390(size_) * size_/rtt_;
                        if (tfrc_psnr_debug_) printf("4444 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
                }
        }
	if (debug_) printf("NO FEEDBACK: time: %5.2f rate: %5.2f all_idle: %d\n", now, rate_, all_idle_);
	UrgentFlag = 1;
	round_id ++ ;
	double t = 2*rtt_ ; 
	// Set the nofeedback timer.
	if (t < 2*size_/rate_) 
		t = 2*size_/rate_ ; 
	NoFeedbacktimer_.resched(t);
	if (datalimited_) {
		all_idle_ = 1;
		if (debug_) printf("Time: %5.2f Datalimited now.\n", now);
	}
	nextpkt();
}

void TfrcPSNRAgent::reduce_rate_on_no_feedback_psnr()
{
	double now = Scheduler::instance().clock();
	if (tfrc_psnr_debug_) printf("At %f reduce rate on no feedback rate_ = %f = %f in kbps\n", now, rate_, rate_*8.0/1000.0);
	if (rate_change_ != SLOW_START)
		// "if" statement added by Sally on 9/25/06,
		// so that an idle report doesn't kick us out of
		// slow-start!
		rate_change_ = RATE_DECREASE;
	if (oldCode_ || (!all_idle_ && !datalimited_)) {
		// if we are not datalimited
		rate_*=0.5;
		if (tfrc_psnr_debug_) printf("1111 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
	} else if ((datalimited_ || all_idle_) && rate_init_option_ == 1) {
		// all_idle_: the sender has been datalimited since the
		//    timer was set
		//  Don't reduce rate below rate_init_ * size_/rtt_.
                if (rate_ > 2.0 * rate_init_ * size_/rtt_ ) {
                        rate_*=0.5;
                        if (tfrc_psnr_debug_) printf("2222 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
                }
	} else if ((datalimited_ || all_idle_) && rate_init_option_ == 2) {
                // Don't reduce rate below the RFC3390 rate.
                if (rate_ > 2.0 * rfc3390(size_) * size_/rtt_ ) {
                        rate_*=0.5;
                        if (tfrc_psnr_debug_) printf("3333 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
                } else if ( rate_ > rfc3390(size_) * size_/rtt_ ) {
                        rate_ = rfc3390(size_) * size_/rtt_;
                        if (tfrc_psnr_debug_) printf("4444 rate_ = %f = %f in kbps\n", rate_, rate_*8.0/1000.0);
                }
        }
	if (debug_) printf("NO FEEDBACK: time: %5.2f rate: %5.2f all_idle: %d\n", now, rate_, all_idle_);
	UrgentFlag = 1;
	round_id ++ ;
	double t = 2*rtt_ ;
	// Set the nofeedback timer.
	if (t < 2*size_/rate_)
		t = 2*size_/rate_ ;
	NoFeedbacktimer_.resched(t);
	if (datalimited_) {
		all_idle_ = 1;
		if (debug_) printf("Time: %5.2f Datalimited now.\n", now);
	}
	nextpkt();
}

void TfrcPSNRSendTimer::expire(Event *) {
  	a_->nextpkt();
}

void TfrcPSNRNoFeedbackTimer::expire(Event *) {
	a_->reduce_rate_on_no_feedback ();
}
